@@ -20,8 +20,6 @@
 #
 
 import logging
-import os
-import subprocess
 from time import sleep
 
 import paramiko
@@ -34,7 +32,7 @@ from libcloud.compute.types import KeyPairDoesNotExistError
 
 import fedimg
 import fedimg.messenger
-from fedimg.util import get_file_arch, ssh_connection_works
+from fedimg.util import get_file_arch, get_virt_type, ssh_connection_works
 
 
 class EC2ServiceException(Exception):
@@ -57,6 +55,15 @@ class EC2Service(object):
     """ A class for interacting with an EC2 connection. """
 
     def __init__(self):
+        self.util_node = None
+        self.util_volume = None
+        self.snapshot = None
+        self.image = None
+        self.test_node = None
+        self.build_name = 'Fedimg build'
+        self.destination = 'somewhere'
+        self.test_success = False
+
         # Will be a list of dicts. Dicts will contain AMI info.
         self.amis = list()
 
@@ -91,40 +98,59 @@ class EC2Service(object):
                      'us-west-2': Provider.EC2_US_WEST_OREGON}
         return providers[region]
 
+    def _clean_up(self, driver, delete_image=False):
+        """ Cleans up resources via a libcloud driver. """
+        logging.info('Cleaning up resources')
+        if delete_image and self.image:
+            driver.delete_image(self.image)
+            self.image = None
+
+        if self.snapshot and not self.image:
+            driver.destroy_volume_snapshot(self.snapshot)
+            self.snapshot = None
+
+        if self.util_node:
+            driver.destroy_node(self.util_node)
+            # Wait for node to be terminated
+            while ssh_connection_works(fedimg.AWS_UTIL_USER,
+                                       self.util_node.public_ips[0],
+                                       fedimg.AWS_KEYPATH):
+                sleep(10)
+            self.util_node = None
+        if self.util_volume:
+            # Destroy /dev/sdb or whatever
+            driver.destroy_volume(self.util_volume)
+            self.util_volume = None
+        if self.test_node:
+            driver.destroy_node(self.test_node)
+            self.test_node = None
+
     def upload(self, raw_url):
         """ Takes a URL to a .raw.xz file and registers it as an AMI in each
         EC2 region. """
 
-        node = None
-        volume = None
-        sda_vol = None
-        snapshot = None
-        image = None
-        test_node = None
-        build_name = 'Fedimg build'
-        destination = 'somewhere'
-        test_success = False
-
         logging.info('EC2 upload process started')
 
-        fedimg.messenger.message('image.upload', build_name, destination,
-                                 'started')
+        fedimg.messenger.message('image.upload', self.build_name,
+                                 self.destination, 'started')
 
         try:
             file_name = raw_url.split('/')[-1]
-            build_name = file_name.replace('.raw.xz', '')
+            self.build_name = file_name.replace('.raw.xz', '')
             image_arch = get_file_arch(file_name)
+            # no EBS-enabled instance types offer a 32 bit architecture
+            self.amis = [a for a in self.amis if a['arch'] == 'x86_64']
             ami = self.amis[0]
-            destination = 'EC2 ({region})'.format(region=ami['region'])
+            self.destination = 'EC2 ({region})'.format(region=ami['region'])
 
             cls = get_driver(ami['prov'])
             driver = cls(fedimg.AWS_ACCESS_ID, fedimg.AWS_SECRET_KEY)
 
             # select the desired node attributes
             sizes = driver.list_sizes()
-            size_id = 'm1.large'
+            reg_size_id = 'm1.large'
             # check to make sure we have access to that size node
-            size = [s for s in sizes if s.id == size_id][0]
+            size = [s for s in sizes if s.id == reg_size_id][0]
             base_image = NodeImage(id=ami['ami'], name=None, driver=driver)
 
             # deploy node
@@ -155,19 +181,21 @@ class EC2Service(object):
             # Must be EBS-backed for AMI registration to work.
             while True:
                 try:
-                    node = driver.deploy_node(name=name, image=base_image,
-                                              size=size,
-                                              ssh_username=fedimg.UTIL_USER,
-                                              ssh_alternate_usernames=[''],
-                                              ssh_key=fedimg.AWS_KEYPATH,
-                                              deploy=msd,
-                                              kernel_id=ami['aki'],
-                                              ex_metadata={'build':
-                                                           build_name},
-                                              ex_keyname=fedimg.AWS_KEYNAME,
-                                              ex_security_groups=['ssh'],
-                                              ex_ebs_optimized=True,
-                                              ex_blockdevicemappings=mappings)
+                    self.util_node = driver.deploy_node(
+                        name=name,
+                        image=base_image,
+                        size=size,
+                        ssh_username=fedimg.AWS_UTIL_USER,
+                        ssh_alternate_usernames=[''],
+                        ssh_key=fedimg.AWS_KEYPATH,
+                        deploy=msd,
+                        kernel_id=ami['aki'],
+                        ex_metadata={'build':
+                                     self.build_name},
+                        ex_keyname=fedimg.AWS_KEYNAME,
+                        ex_security_groups=['ssh'],
+                        ex_ebs_optimized=True,
+                        ex_blockdevicemappings=mappings)
 
                 except KeyPairDoesNotExistError:
                     # The keypair is missing from the current region.
@@ -196,8 +224,8 @@ class EC2Service(object):
                 break
 
             # Wait until the utility node has SSH running
-            while not ssh_connection_works(fedimg.UTIL_USER,
-                                           node.public_ips[0],
+            while not ssh_connection_works(fedimg.AWS_UTIL_USER,
+                                           self.util_node.public_ips[0],
                                            fedimg.AWS_KEYPATH):
                 sleep(10)
 
@@ -205,7 +233,8 @@ class EC2Service(object):
 
             client = paramiko.SSHClient()
             client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
-            client.connect(node.public_ips[0], username=fedimg.UTIL_USER,
+            client.connect(self.util_node.public_ips[0],
+                           username=fedimg.AWS_UTIL_USER,
                            key_filename=fedimg.AWS_KEYPATH)
             cmd = "sudo sh -c 'curl {0} | xzcat > /dev/xvdb'".format(raw_url)
             chan = client.get_transport().open_session()
@@ -227,16 +256,17 @@ class EC2Service(object):
 
             # Get volume name that image was written to
             vol_id = [x['ebs']['volume_id'] for x in
-                      node.extra['block_device_mapping'] if
+                      self.util_node.extra['block_device_mapping'] if
                       x['device_name'] == '/dev/sdb'][0]
 
             logging.info('Destroying utility node')
 
             # Terminate the utility instance
-            driver.destroy_node(node)
+            driver.destroy_node(self.util_node)
 
             # Wait for utility node to be terminated
-            while ssh_connection_works(fedimg.UTIL_USER, node.public_ips[0],
+            while ssh_connection_works(fedimg.AWS_UTIL_USER,
+                                       self.util_node.public_ips[0],
                                        fedimg.AWS_KEYPATH):
                 sleep(10)
 
@@ -246,62 +276,81 @@ class EC2Service(object):
             sleep(45)
 
             # Take a snapshot of the volume the image was written to
-            volume = [v for v in driver.list_volumes() if v.id == vol_id][0]
-            snap_name = 'fedimg-snap-{0}'.format(build_name)
+            self.util_volume = [v for v in driver.list_volumes()
+                                if v.id == vol_id][0]
+            snap_name = 'fedimg-snap-{0}'.format(self.build_name)
 
             logging.info('Taking a snapshot of the written volume')
 
-            snapshot = driver.create_volume_snapshot(volume,
-                                                     name=snap_name)
-            snap_id = str(snapshot.id)
+            self.snapshot = driver.create_volume_snapshot(self.util_volume,
+                                                          name=snap_name)
+            snap_id = str(self.snapshot.id)
 
-            while snapshot.extra['state'] != 'completed':
+            while self.snapshot.extra['state'] != 'completed':
                 # need to re-pull snapshot object to get updates
-                snapshot = [s for s in driver.list_snapshots()
-                            if s.id == snap_id][0]
+                self.snapshot = [s for s in driver.list_snapshots()
+                                 if s.id == snap_id][0]
                 sleep(10)
 
             logging.info('Snapshot taken')
 
             # Delete the volume now that we've got the snapshot
-            driver.destroy_volume(volume)
-            volume = None  # make sure Fedimg knows that the vol is gone
+            driver.destroy_volume(self.util_volume)
+            # make sure Fedimg knows that the vol is gone
+            self.util_volume = None
 
             logging.info('Destroyed volume')
 
+            # Actually register image
+            logging.info('Registering image as an AMI')
+            image_name = "{0}-{1}-0".format(self.build_name, ami['region'])
+
+            virt_type = get_virt_type(image_name)
+
+            if virt_type == 'paravirtual':
+                test_size_id = 'm1.medium'
+                registration_aki = ami['aki']
+                test_aki = ami['aki']
+                reg_root_device_name = '/dev/sda'
+            else:  # HVM
+                test_size_id = 'm3.medium'
+                # Can't supply a kernel image with HVM
+                registration_aki = None
+                test_aki = None
+                reg_root_device_name = '/dev/sda1'
+
             # Block device mapping for the AMI
-            mapping = [{'DeviceName': '/dev/sda',
+            mapping = [{'DeviceName': reg_root_device_name,
                         'Ebs': {'SnapshotId': snap_id,
                                 'VolumeSize': 12,
                                 'VolumeType': 'standard',
                                 'DeleteOnTermination': 'true'}}]
 
-            # Actually register image
-            logging.info('Registering image as an AMI')
-            image_name = "{0}-{1}".format(build_name, ami['region'])
-            # Avoid duplicate image name by adding a '-' and a number to the
-            # end if there is already an AMI with that name.
+            # Avoid duplicate image name by incrementing the number at the
+            # end of the image name if there is already an AMI with that name.
             dup_count = 0  # counter: number of AMIs with same base image name
             while True:
                 try:
-                    if dup_count == 1:
-                        image_name += '-1'  # avoid duplicate image name
-                    elif dup_count > 1:
-                        # Remove trailing '-1' or '-2' or '-3' or...
+                    if dup_count > 0:
+                        # Remove trailing '-0' or '-1' or '-2' or...
                         image_name = '-'.join(image_name.split('-')[:-1])
                         # Re-add trailing dup number with new count
                         image_name += '-{0}'.format(dup_count)
-                    image = driver.ex_register_image(
+                    # Try to register with that name
+                    self.image = driver.ex_register_image(
                         image_name,
                         description=None,
-                        root_device_name='/dev/sda',
+                        root_device_name=reg_root_device_name,
                         block_device_mapping=mapping,
-                        kernel_id=ami['aki'],
+                        virtualization_type=virt_type,
+                        kernel_id=registration_aki,
                         architecture=image_arch)
                 except Exception as e:
+                    # Check if the problem was a duplicate name
                     if 'InvalidAMIName.Duplicate' in e.message:
+                        # Keep trying until an unused name is found
                         dup_count += 1
-                        continue  # Keep trying until an unused name is found
+                        continue
                     else:
                         raise
                 break
@@ -309,8 +358,8 @@ class EC2Service(object):
             logging.info('Completed image registration')
 
             # Emit success fedmsg
-            fedimg.messenger.message('image.upload', build_name, destination,
-                                     'completed')
+            fedimg.messenger.message('image.upload', self.build_name,
+                                     self.destination, 'completed')
 
             # Spin up a node of the AMI to test
 
@@ -322,41 +371,39 @@ class EC2Service(object):
             # Create deployment object
             msd = MultiStepDeployment([step_1, step_2])
 
-            mappings = [{'VirtualName': None,  # cannot specify with Ebs
-                         'Ebs': {'VolumeSize': 12,  # 12 GB should be enough
-                                 'VolumeType': 'standard',
-                                 'DeleteOnTermination': 'true'},
-                         'DeviceName': '/dev/sda'}]
-
             logging.info('Deploying test node')
 
             name = 'Fedimg AMI tester'
-            test_node = driver.deploy_node(name=name, image=image, size=size,
-                                           ssh_username=fedimg.TEST_USER,
-                                           ssh_alternate_usernames=['root'],
-                                           ssh_key=fedimg.AWS_KEYPATH,
-                                           deploy=msd,
-                                           kernel_id=ami['aki'],
-                                           ex_metadata={'build': build_name},
-                                           ex_keyname=fedimg.AWS_KEYNAME,
-                                           ex_security_groups=['ssh'],
-                                           ex_ebs_optimized=True)
+            size = [s for s in sizes if s.id == test_size_id][0]
+
+            self.test_node = driver.deploy_node(
+                name=name, image=self.image, size=size,
+                ssh_username=fedimg.AWS_TEST_USER,
+                ssh_alternate_usernames=['root'],
+                ssh_key=fedimg.AWS_KEYPATH,
+                deploy=msd,
+                kernel_id=test_aki,
+                ex_metadata={'build': self.build_name},
+                ex_keyname=fedimg.AWS_KEYNAME,
+                ex_security_groups=['ssh'],
+                )
 
             # Wait until the test node has SSH running
-            while not ssh_connection_works(fedimg.TEST_USER,
-                                           test_node.public_ips[0],
+            while not ssh_connection_works(fedimg.AWS_TEST_USER,
+                                           self.test_node.public_ips[0],
                                            fedimg.AWS_KEYPATH):
                 sleep(10)
 
             logging.info('Starting AMI tests')
 
             # Alert the fedmsg bus that an image test has started
-            fedimg.messenger.message('image.test', build_name, destination,
-                                     'started')
+            fedimg.messenger.message('image.test', self.build_name,
+                                     self.destination, 'started')
 
             client = paramiko.SSHClient()
             client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
-            client.connect(test_node.public_ips[0], username=fedimg.TEST_USER,
+            client.connect(self.test_node.public_ips[0],
+                           username=fedimg.AWS_TEST_USER,
                            key_filename=fedimg.AWS_KEYPATH)
             cmd = "true"
             chan = client.get_transport().open_session()
@@ -371,99 +418,101 @@ class EC2Service(object):
                 raise EC2AMITestException("Tests on AMI failed.")
 
             logging.info('AMI test completed')
-            fedimg.messenger.message('image.test', build_name, destination,
-                                     'completed')
-            test_success = True
+            fedimg.messenger.message('image.test', self.build_name,
+                                     self.destination, 'completed')
+            self.test_success = True
 
             logging.info('Destroying test node')
 
             # Destroy the test node
-            driver.destroy_node(test_node)
+            driver.destroy_node(self.test_node)
+
+            # Make AMI public
+            driver.ex_modify_image_attribute(
+                self.image,
+                {'LaunchPermission.Add.1.Group': 'all'})
 
         except EC2UtilityException as e:
-            fedimg.messenger.message('image.upload', build_name, destination,
-                                     'failed')
+            fedimg.messenger.message('image.upload', self.build_name,
+                                     self.destination, 'failed')
             print "Failure:", e
+            if fedimg.CLEAN_UP_ON_FAILURE:
+                self._clean_up(driver,
+                               delete_image=fedimg.DELETE_IMAGE_ON_FAILURE)
 
         except EC2AMITestException as e:
-            fedimg.messenger.message('image.test', build_name, destination,
-                                     'failed')
+            fedimg.messenger.message('image.test', self.build_name,
+                                     self.destination, 'failed')
             print "Failure:", e
+            if fedimg.CLEAN_UP_ON_FAILURE:
+                self._clean_up(driver,
+                               delete_image=fedimg.DELETE_IMAGE_ON_FAILURE)
 
         except DeploymentException as e:
-            fedimg.messenger.message('image.upload', build_name, destination,
-                                     'failed')
+            fedimg.messenger.message('image.upload', self.build_name,
+                                     self.destination, 'failed')
             print "Problem deploying node: {0}".format(e.value)
-            print "Terminating instance."
-            driver.destroy_node(e.node)
+            if fedimg.CLEAN_UP_ON_FAILURE:
+                self._clean_up(driver,
+                               delete_image=fedimg.DELETE_IMAGE_ON_FAILURE)
 
         except Exception as e:
             # Just give a general failure message.
-            fedimg.messenger.message('image.upload', build_name, destination,
-                                     'failed')
+            fedimg.messenger.message('image.upload', self.build_name,
+                                     self.destination, 'failed')
             print "Unexpected exception:", e
-            print "Terminating instance and destroying other resources."
-            if sda_vol:
-                driver.destroy_volume(sda_vol)
-            if snapshot:
-                if image:
-                    driver.delete_image(image)
-                    driver.destroy_volume_snapshot(snapshot)
+            if fedimg.CLEAN_UP_ON_FAILURE:
+                self._clean_up(driver,
+                               delete_image=fedimg.DELETE_IMAGE_ON_FAILURE)
 
-        finally:
-            logging.info('Cleaning up resources')
-            if node:
-                driver.destroy_node(node)
-                # Wait for node to be terminated
-                while ssh_connection_works('ec2_user', node.public_ips[0],
-                                           fedimg.AWS_KEYPATH):
-                    sleep(10)
-            if sda_vol:
-                # Destroy /dev/sda volume if lagging behind
-                driver.destroy_volume(sda_vol)
-            if volume:
-                # Destroy /dev/sdb or whatever
-                driver.destroy_volume(volume)
-            if test_node:
-                driver.destroy_node(test_node)
+        else:
+            self._clean_up(driver)
 
-        if test_success:
+        if self.test_success:
             # Copy the AMI to every other region if tests passed
+            copied_images = list()  # completed image copies (ami: image)
             for ami in self.amis[1:]:
-                # Avoid duplicate image name by adding a '-' and a number to
-                # the end if there is already an AMI with that name.
+
+                alt_dest = 'EC2 ({region})'.format(
+                    region=ami['region'])
+
+                fedimg.messenger.message('image.upload',
+                                         self.build_name,
+                                         alt_dest, 'started')
+
+                alt_cls = get_driver(ami['prov'])
+                alt_driver = alt_cls(fedimg.AWS_ACCESS_ID,
+                                     fedimg.AWS_SECRET_KEY)
+
+                image_name = "{0}-{1}-0".format(
+                    self.build_name, ami['region'])
+
+                logging.info('AMI copy to {0} started'.format(
+                    ami['region']))
+
+                # Avoid duplicate image name by incrementing the number at the
+                # end of the image name if there is already an AMI with
+                # that name.
                 dup_count = 0  # counter: num of AMIs with same base image name
                 while True:
                     try:
-                        if dup_count == 1:
-                            image_name += '-1'  # avoid duplicate image name
-                        elif dup_count > 1:
-                            # Remove trailing '-1' or '-2' or '-3' or...
+                        if dup_count > 0:
+                            # Remove trailing '-0' or '-1' or '-2' or...
                             image_name = '-'.join(image_name.split('-')[:-1])
                             # Re-add trailing dup number with new count
                             image_name += '-{0}'.format(dup_count)
-                        alt_dest = 'EC2 ({region})'.format(
-                            region=ami['region'])
 
-                        fedimg.messenger.message('image.upload', build_name,
-                                                 alt_dest, 'started')
+                        image_copy = alt_driver.copy_image(
+                            self.image,
+                            self.amis[0]['region'],
+                            name=image_name)
 
-                        alt_cls = get_driver(ami['prov'])
-                        alt_driver = alt_cls(fedimg.AWS_ACCESS_ID,
-                                             fedimg.AWS_SECRET_KEY)
+                        copied_images.append(image_copy)
 
-                        image_name = "{0}-{1}".format(
-                            build_name, ami['region'])
-
-                        logging.info('AMI copy to {0} started'.format(
-                            ami['region']))
-
-                        alt_driver.copy_image(image, self.amis[0]['region'],
-                                              name=image_name)
-
-                        fedimg.messenger.message('image.upload', build_name,
-                                                 alt_dest, 'completed')
+                        logging.info('AMI {0} copied to AMI {1}'.format(
+                            self.image, image_name))
                     except Exception as e:
+                        # Check if the problem was a duplicate name
                         if 'InvalidAMIName.Duplicate' in e.message:
                             # Keep trying until an unused name is found
                             dup_count += 1
@@ -474,6 +523,34 @@ class EC2Service(object):
                                 'Image copy to {0} failed'.format(
                                     ami['region']))
                             fedimg.messenger.message('image.upload',
-                                                     build_name,
+                                                     self.build_name,
                                                      alt_dest, 'failed')
                     break
+
+            # Now cycle through and make all of the copied AMIs public
+            # once the copy process has completed.
+            amis = self.amis[1:]
+            for image in copied_images:
+                ami = amis[copied_images.index(image)]
+                alt_cls = get_driver(ami['prov'])
+                alt_driver = alt_cls(fedimg.AWS_ACCESS_ID,
+                                     fedimg.AWS_SECRET_KEY)
+                # Need to wait until the copy finishes in order to make
+                # the AMI public.
+                while True:
+                    try:
+                        alt_driver.ex_modify_image_attribute(
+                            image,
+                            {'LaunchPermission.Add.1.Group': 'all'})
+                    except Exception as e:
+                        if 'InvalidAMIID.Unavailable' in e.message:
+                            sleep(20)
+                            continue
+                    break
+
+                # TODO: Add logging here
+                logging.info('Made image {0} public'.format(image))
+
+                fedimg.messenger.message('image.upload',
+                                         self.build_name,
+                                         alt_dest, 'completed')
