@@ -44,8 +44,8 @@ def get_file_arch(file_name):
 
 
 def get_rawxz_url(task_result):
-    """ Returns the URL of the raw.xz file produced by the Koji
-    task ID represented by the task_result argument. """
+    """ Returns the URL of the raw.xz file produced by the Koji task whose
+    output files are passed as a list via the task_result argument. """
     # I think there might only ever be one qcow2 file per task,
     # but doing it this way plays it safe.
     file_name = [f for f in task_result['files'] if f.endswith('.raw.xz')][0]
@@ -58,6 +58,19 @@ def get_rawxz_url(task_result):
     return full_file_location + "/{}".format(file_name)
 
 
+def get_virt_type(file_name):
+    """ Takes a file name (probably of a .raw.xz image file) and returns
+    the suspected virtualization type that the image file should be
+    registered as. """
+    file_name = file_name.lower()
+    if file_name.find('atomic') != -1:
+        # hvm is required for atomic images
+        return 'hvm'
+    else:
+        # otherwise, use the default (paravirtual)
+        return 'paravirtual'
+
+
 def ssh_connection_works(username, ip, keypath):
     """ Returns True if an SSH connection can me made to `username`@`ip`. """
     ssh = paramiko.SSHClient()
